@@ -157,6 +157,7 @@ enum opcode {
   GEN_OPCODE_FBH = 75,
   GEN_OPCODE_FBL = 76,
   GEN_OPCODE_ADDC = 78,
+  GEN_OPCODE_SUBB = 79,
   GEN_OPCODE_SAD2 = 80,
   GEN_OPCODE_SADA2 = 81,
   GEN_OPCODE_DP4 = 84,
@@ -242,6 +243,8 @@ enum GenMessageTarget {
 #define GEN_TYPE_V   6 /* packed int vector, immediates only, uword dest only */
 #define GEN_TYPE_DF  6
 #define GEN_TYPE_F   7
+#define GEN_TYPE_UL  8
+#define GEN_TYPE_L   9
 
 #define GEN_ARF_NULL                  0x00
 #define GEN_ARF_ADDRESS               0x10
@@ -316,6 +319,15 @@ enum GenMessageTarget {
 #define GEN_BYTE_SCATTER          12//1100: Byte Scattered Write
 #define GEN_UNTYPED_WRITE         13//1101: Untyped Surface Write
 
+/* Data port data cache scratch messages*/
+#define GEN_SCRATCH_READ                  0
+#define GEN_SCRATCH_WRITE                 1
+#define GEN_SCRATCH_CHANNEL_MODE_OWORD    0
+#define GEN_SCRATCH_CHANNEL_MODE_DWORD    1
+#define GEN_SCRATCH_BLOCK_SIZE_1          0
+#define GEN_SCRATCH_BLOCK_SIZE_2          1
+#define GEN_SCRATCH_BLOCK_SIZE_4          3
+
 /* Data port render cache Message Type*/
 #define GEN_MBLOCK_READ           4  //0100: Media Block Read
 #define GEN_TYPED_READ            5  //0101: Typed Surface Read
@@ -331,6 +343,10 @@ enum GenMessageTarget {
 #define GEN_BYTE_SCATTER_DWORD  2
 #define GEN_BYTE_SCATTER_QWORD  3
 
+/* dword scattered rw */
+#define GEN_DWORD_SCATTER_8_DWORDS   2
+#define GEN_DWORD_SCATTER_16_DWORDS  3
+
 #define GEN_SAMPLER_RETURN_FORMAT_FLOAT32     0
 #define GEN_SAMPLER_RETURN_FORMAT_UINT32      2
 #define GEN_SAMPLER_RETURN_FORMAT_SINT32      3
@@ -762,6 +778,22 @@ struct GenInstruction
       uint32_t end_of_thread:1;
     } gen7_byte_rw;
 
+    /*! Data port Scratch Read/ write */
+    struct {
+      uint32_t offset:12;
+      uint32_t block_size:2;
+      uint32_t ignored0:1;
+      uint32_t invalidate_after_read:1;
+      uint32_t channel_mode:1;
+      uint32_t msg_type:1;
+      uint32_t category:1;
+      uint32_t header_present:1;
+      uint32_t response_length:5;
+      uint32_t msg_length:4;
+      uint32_t pad2:2;
+      uint32_t end_of_thread:1;
+    } gen7_scratch_rw;
+
     /*! Data port OBlock read / write */
     struct {
       uint32_t bti:8;
@@ -777,6 +809,21 @@ struct GenInstruction
       uint32_t end_of_thread:1;
     } gen7_oblock_rw;
 
+    /*! Data port dword scatter / gather */
+    struct {
+      uint32_t bti:8;
+      uint32_t block_size:2;
+      uint32_t ignored0:3;
+      uint32_t invalidate_after_read:1;
+      uint32_t msg_type:4;
+      uint32_t ignored1:1;
+      uint32_t header_present:1;
+      uint32_t response_length:5;
+      uint32_t msg_length:4;
+      uint32_t pad2:2;
+      uint32_t end_of_thread:1;
+    } gen7_dword_rw;
+
     /*! Data port typed read / write messages */
     struct {
       uint32_t bti:8;
