@@ -198,7 +198,21 @@ namespace gbe
      insn->bits3.gen7_typed_rw.bti = bti;
      insn->bits3.gen7_typed_rw.msg_type = msg_type;
   }
-
+  static void setDWordScatterMessgae(GenEncoder *p,
+                                     GenInstruction *insn,
+                                     uint32_t bti,
+                                     uint32_t block_size,
+                                     uint32_t msg_type,
+                                     uint32_t msg_length,
+                                     uint32_t response_length)
+  {
+    const GenMessageTarget sfid = GEN6_SFID_DATAPORT_CONSTANT_CACHE;
+    setMessageDescriptor(p, insn, sfid, msg_length, response_length);
+    insn->bits3.gen7_dword_rw.msg_type = msg_type;
+    insn->bits3.gen7_dword_rw.bti = bti;
+    insn->bits3.gen7_dword_rw.block_size = block_size;
+    insn->bits3.gen7_dword_rw.invalidate_after_read = 0;
+  }
   //////////////////////////////////////////////////////////////////////////
   // Gen Emitter encoding class
   //////////////////////////////////////////////////////////////////////////
@@ -315,6 +329,7 @@ namespace gbe
 
   void GenEncoder::setSrc1(GenInstruction *insn, GenRegister reg) {
      assert(reg.nr < 128);
+     assert(reg.file != GEN_ARCHITECTURE_REGISTER_FILE || reg.nr == 0);
 
      insn->bits1.da1.src1_reg_file = reg.file;
      insn->bits1.da1.src1_reg_type = reg.type;
@@ -356,103 +371,74 @@ namespace gbe
     0
   };
 
-  static int dst_type(int exec_width) {
-    if (exec_width == 8)
-      return GEN_TYPE_UD;
-    if (exec_width == 16)
-      return GEN_TYPE_UW;
-    NOT_IMPLEMENTED;
-    return 0;
-  }
-
-  void GenEncoder::READ_FLOAT64(GenRegister dst, GenRegister src, uint32_t bti, uint32_t elemNum) {
-    int w = curr.execWidth;
-    dst = GenRegister::h2(dst);
-    dst.type = GEN_TYPE_UD;
-    src.type = GEN_TYPE_UD;
-    GenRegister r = GenRegister::retype(GenRegister::suboffset(src, w*2), GEN_TYPE_UD);
-    GenRegister imm4 = GenRegister::immud(4);
-    GenInstruction *insn;
-    insn = next(GEN_OPCODE_SEND);
-    setHeader(insn);
-    setDst(insn, GenRegister::uw16grf(r.nr, 0));
-    setSrc0(insn, GenRegister::ud8grf(src.nr, 0));
-    setSrc1(insn, GenRegister::immud(0));
-    setDPUntypedRW(this, insn, bti, untypedRWMask[1], GEN_UNTYPED_READ, curr.execWidth / 8, curr.execWidth / 8);
+  void GenEncoder::READ64(GenRegister dst, GenRegister tmp, GenRegister addr, GenRegister src, uint32_t bti, uint32_t elemNum) {
+    GenRegister dst32 = GenRegister::retype(dst, GEN_TYPE_UD);
+    src = GenRegister::retype(src, GEN_TYPE_UD);
+    addr = GenRegister::retype(addr, GEN_TYPE_UD);
+    tmp = GenRegister::retype(tmp, GEN_TYPE_UD);
+    uint32_t originSimdWidth = curr.execWidth;
+    uint32_t originPredicate = curr.predicate;
+    uint32_t originMask = curr.noMask;
     push();
-    curr.quarterControl = 0;
-    curr.nibControl = 0;
-    MOV(dst, r);
-    if (w == 8)
-      curr.nibControl = 1;
-    else
-      curr.quarterControl = 1;
-    MOV(GenRegister::suboffset(dst, w), GenRegister::suboffset(r, w / 2));
-    pop();
-    ADD(src, src, imm4);
-    insn = next(GEN_OPCODE_SEND);
-    setHeader(insn);
-    setDst(insn, GenRegister::uw16grf(r.nr, 0));
-    setSrc0(insn, GenRegister::ud8grf(src.nr, 0));
-    setSrc1(insn, GenRegister::immud(0));
-    setDPUntypedRW(this, insn, bti, untypedRWMask[1], GEN_UNTYPED_READ, curr.execWidth / 8, curr.execWidth / 8);
-    push();
-    curr.quarterControl = 0;
-    curr.nibControl = 0;
-    MOV(GenRegister::suboffset(dst, 1), r);
-    if (w == 8)
-      curr.nibControl = 1;
-    else
-      curr.quarterControl = 1;
-    MOV(GenRegister::suboffset(dst, w + 1), GenRegister::suboffset(r, w / 2));
+    for ( uint32_t channels = 0, currQuarter = GEN_COMPRESSION_Q1;
+          channels < originSimdWidth; channels += 8, currQuarter++) {
+      curr.predicate = GEN_PREDICATE_NONE;
+      curr.noMask = GEN_MASK_DISABLE;
+      curr.execWidth = 8;
+      /* XXX The following instruction is illegal, but it works as SIMD 1*4 mode
+         which is what we want here. */
+      MOV(GenRegister::h2(addr), GenRegister::suboffset(src, channels));
+      ADD(GenRegister::h2(GenRegister::suboffset(addr, 1)), GenRegister::suboffset(src, channels), GenRegister::immd(4));
+      MOV(GenRegister::h2(GenRegister::suboffset(addr, 8)), GenRegister::suboffset(src, channels + 4));
+      ADD(GenRegister::h2(GenRegister::suboffset(addr, 9)), GenRegister::suboffset(src, channels + 4), GenRegister::immd(4));
+      // Let's use SIMD16 to read all bytes for 8 doubles data at one time.
+      curr.execWidth = 16;
+      this->UNTYPED_READ(tmp, addr, bti, elemNum);
+      if (originSimdWidth == 16)
+        curr.quarterControl = currQuarter;
+      curr.predicate = originPredicate;
+      curr.noMask = originMask;
+      // Back to simd8 for correct predication flag.
+      curr.execWidth = 8;
+      MOV(GenRegister::retype(GenRegister::suboffset(dst32, channels * 2), GEN_TYPE_DF), GenRegister::retype(tmp, GEN_TYPE_DF));
+    }
     pop();
   }
 
-  void GenEncoder::WRITE_FLOAT64(GenRegister msg, uint32_t bti, uint32_t elemNum) {
-    int w = curr.execWidth;
-    GenRegister r = GenRegister::retype(GenRegister::suboffset(msg, w*3), GEN_TYPE_UD);
-    r.type = GEN_TYPE_UD;
-    GenRegister hdr = GenRegister::h2(r);
-    GenRegister src = GenRegister::ud16grf(msg.nr + w / 8, 0);
-    src.hstride = GEN_HORIZONTAL_STRIDE_2;
-    GenRegister data = GenRegister::offset(r, w / 8);
-    GenRegister imm4 = GenRegister::immud(4);
-    MOV(r, GenRegister::ud8grf(msg.nr, 0));
+  void GenEncoder::WRITE64(GenRegister msg, GenRegister data, uint32_t bti, uint32_t elemNum, bool is_scalar) {
+    GenRegister data32 = GenRegister::retype(data, GEN_TYPE_UD);
+    GenRegister unpacked;
+    msg = GenRegister::retype(msg, GEN_TYPE_UD);
+    int originSimdWidth = curr.execWidth;
+    int originPredicate = curr.predicate;
+    int originMask = curr.noMask;
     push();
-    curr.quarterControl = 0;
-    curr.nibControl = 0;
-    MOV(data, src);
-    if (w == 8)
-      curr.nibControl = 1;
-    else
-      curr.quarterControl = 1;
-    MOV(GenRegister::suboffset(data, w / 2), GenRegister::suboffset(src, w));
+    for (uint32_t half = 0; half < 2; half++) {
+      curr.predicate = GEN_PREDICATE_NONE;
+      curr.noMask = GEN_MASK_DISABLE;
+      curr.execWidth = 8;
+      if (is_scalar) {
+        unpacked = data32;
+        unpacked.subnr += half * 4;
+      } else
+        unpacked = GenRegister::unpacked_ud(data32.nr, data32.subnr + half);
+      MOV(GenRegister::suboffset(msg, originSimdWidth), unpacked);
+      if (originSimdWidth == 16) {
+        if (is_scalar) {
+          unpacked = data32;
+          unpacked.subnr += half * 4;
+        } else
+          unpacked = GenRegister::unpacked_ud(data32.nr + 2, data32.subnr + half);
+        MOV(GenRegister::suboffset(msg, originSimdWidth + 8), unpacked);
+        curr.execWidth = 16;
+      }
+      if (half == 1)
+        ADD(GenRegister::retype(msg, GEN_TYPE_UD), GenRegister::retype(msg, GEN_TYPE_UD), GenRegister::immd(4));
+      curr.predicate = originPredicate;
+      curr.noMask = originMask;
+      this->UNTYPED_WRITE(msg, bti, elemNum);
+    }
     pop();
-    GenInstruction *insn;
-    insn = next(GEN_OPCODE_SEND);
-    setHeader(insn);
-    setDst(insn, GenRegister::retype(GenRegister::null(), dst_type(curr.execWidth)));
-    setSrc0(insn, GenRegister::ud8grf(hdr.nr, 0));
-    setSrc1(insn, GenRegister::immud(0));
-    setDPUntypedRW(this, insn, bti, untypedRWMask[1], GEN_UNTYPED_WRITE, curr.execWidth / 4, 0);
-
-    ADD(r, GenRegister::ud8grf(msg.nr, 0), imm4);
-    push();
-    curr.quarterControl = 0;
-    curr.nibControl = 0;
-    MOV(data, GenRegister::suboffset(src, 1));
-    if (w == 8)
-      curr.nibControl = 1;
-    else
-      curr.quarterControl = 1;
-    MOV(GenRegister::suboffset(data, w / 2), GenRegister::suboffset(src, w + 1));
-    pop();
-    insn = next(GEN_OPCODE_SEND);
-    setHeader(insn);
-    setDst(insn, GenRegister::retype(GenRegister::null(), dst_type(curr.execWidth)));
-    setSrc0(insn, GenRegister::ud8grf(hdr.nr, 0));
-    setSrc1(insn, GenRegister::immud(0));
-    setDPUntypedRW(this, insn, bti, untypedRWMask[1], GEN_UNTYPED_WRITE, curr.execWidth / 4, 0);
   }
 
   void GenEncoder::UNTYPED_READ(GenRegister dst, GenRegister src, uint32_t bti, uint32_t elemNum) {
@@ -470,7 +456,7 @@ namespace gbe
       NOT_IMPLEMENTED;
 
     this->setHeader(insn);
-    this->setDst(insn, GenRegister::uw16grf(dst.nr, 0));
+    this->setDst(insn,  GenRegister::uw16grf(dst.nr, 0));
     this->setSrc0(insn, GenRegister::ud8grf(src.nr, 0));
     this->setSrc1(insn, GenRegister::immud(0));
     setDPUntypedRW(this,
@@ -558,6 +544,36 @@ namespace gbe
                            response_length);
   }
 
+  void GenEncoder::DWORD_GATHER(GenRegister dst, GenRegister src, uint32_t bti) {
+    GenInstruction *insn = this->next(GEN_OPCODE_SEND);
+    uint32_t msg_length = 0;
+    uint32_t response_length = 0;
+    uint32_t block_size = 0;
+    if (this->curr.execWidth == 8) {
+      msg_length = 1;
+      response_length = 1;
+      block_size = GEN_DWORD_SCATTER_8_DWORDS;
+    } else if (this->curr.execWidth == 16) {
+      msg_length = 2;
+      response_length = 2;
+      block_size = GEN_DWORD_SCATTER_16_DWORDS;
+    } else
+      NOT_IMPLEMENTED;
+
+    this->setHeader(insn);
+    this->setDst(insn, dst);
+    this->setSrc0(insn, src);
+    this->setSrc1(insn, GenRegister::immud(0));
+    setDWordScatterMessgae(this,
+                           insn,
+                           bti,
+                           block_size,
+                           GEN_DWORD_GATHER,
+                           msg_length,
+                           response_length);
+
+  }
+
   void GenEncoder::ATOMIC(GenRegister dst, uint32_t function, GenRegister src, uint32_t bti, uint32_t srcNum) {
     GenInstruction *insn = this->next(GEN_OPCODE_SEND);
     uint32_t msg_length = 0;
@@ -601,24 +617,64 @@ namespace gbe
      return &this->store.back();
   }
 
-  INLINE void alu1(GenEncoder *p, uint32_t opcode, GenRegister dst, GenRegister src) {
-     if (dst.isdf() && src.isdf()) {
+  INLINE void _handleDouble(GenEncoder *p, uint32_t opcode, GenRegister dst,
+                            GenRegister src0, GenRegister src1 = GenRegister::null()) {
        int w = p->curr.execWidth;
        p->push();
-       p->curr.quarterControl = 0;
        p->curr.nibControl = 0;
        GenInstruction *insn = p->next(opcode);
        p->setHeader(insn);
        p->setDst(insn, dst);
-       p->setSrc0(insn, src);
+       p->setSrc0(insn, src0);
+       if (!GenRegister::isNull(src1))
+         p->setSrc1(insn, src1);
        if (w == 8)
          p->curr.nibControl = 1; // second 1/8 mask
-       else // w == 16
-         p->curr.quarterControl = 1; // second 1/4 mask
        insn = p->next(opcode);
        p->setHeader(insn);
        p->setDst(insn, GenRegister::suboffset(dst, w / 2));
-       p->setSrc0(insn, GenRegister::suboffset(src, w / 2));
+       p->setSrc0(insn, GenRegister::suboffset(src0, w / 2));
+       if (!GenRegister::isNull(src1))
+         p->setSrc1(insn, GenRegister::suboffset(src1, w / 2));
+       p->pop();
+  }
+
+  // Double register accessing is a little special,
+  // Per Gen spec, then only supported mode is SIMD8 and, it only
+  // handles four doubles each time.
+  // We need to lower down SIMD16 to two SIMD8 and lower down SIMD8
+  // to two SIMD1x4.
+  INLINE void handleDouble(GenEncoder *p, uint32_t opcode, GenRegister dst,
+                           GenRegister src0, GenRegister src1 = GenRegister::null()) {
+      if (p->curr.execWidth == 8)
+        _handleDouble(p, opcode, dst, src0, src1);
+      else if (p->curr.execWidth == 16) {
+        p->push();
+        p->curr.execWidth = 8;
+        p->curr.quarterControl = GEN_COMPRESSION_Q1;
+        _handleDouble(p, opcode, dst, src0, src1);
+        p->curr.quarterControl = GEN_COMPRESSION_Q2;
+        if (!GenRegister::isNull(src1))
+          src1 = GenRegister::offset(src1, 2);
+        _handleDouble(p, opcode, GenRegister::offset(dst, 2), GenRegister::offset(src0, 2), src1);
+        p->pop();
+      }
+  }
+
+  INLINE void alu1(GenEncoder *p, uint32_t opcode, GenRegister dst, GenRegister src) {
+     if (dst.isdf() && src.isdf()) {
+       handleDouble(p, opcode, dst, src);
+     } else if (dst.isint64() && src.isint64()) { // handle int64
+       int execWidth = p->curr.execWidth;
+       p->push();
+       p->curr.execWidth = 8;
+       for (int nib = 0; nib < execWidth / 4; nib ++) {
+         p->curr.chooseNib(nib);
+         p->MOV(dst.bottom_half(), src.bottom_half());
+         p->MOV(dst.top_half(), src.top_half());
+         dst = GenRegister::suboffset(dst, 4);
+         src = GenRegister::suboffset(src, 4);
+       }
        p->pop();
      } else if (needToSplitAlu1(p, dst, src) == false) {
        GenInstruction *insn = p->next(opcode);
@@ -653,25 +709,7 @@ namespace gbe
                    GenRegister src1)
   {
     if (dst.isdf() && src0.isdf() && src1.isdf()) {
-       int w = p->curr.execWidth;
-       p->push();
-       p->curr.quarterControl = 0;
-       p->curr.nibControl = 0;
-       GenInstruction *insn = p->next(opcode);
-       p->setHeader(insn);
-       p->setDst(insn, dst);
-       p->setSrc0(insn, src0);
-       p->setSrc1(insn, src1);
-       if (w == 8)
-         p->curr.nibControl = 1; // second 1/8 mask
-       else // w == 16
-         p->curr.quarterControl = 1; // second 1/4 mask
-       insn = p->next(opcode);
-       p->setHeader(insn);
-       p->setDst(insn, GenRegister::suboffset(dst, w / 2));
-       p->setSrc0(insn, GenRegister::suboffset(src0, w / 2));
-       p->setSrc1(insn, GenRegister::suboffset(src1, w / 2));
-       p->pop();
+       handleDouble(p, opcode, dst, src0, src1);
     } else if (needToSplitAlu2(p, dst, src0, src1) == false) {
        GenInstruction *insn = p->next(opcode);
        p->setHeader(insn);
@@ -808,7 +846,16 @@ namespace gbe
     r.width = GEN_WIDTH_1;
     r.hstride = GEN_HORIZONTAL_STRIDE_0;
     push();
+    uint32_t width = curr.execWidth;
+    curr.execWidth = 8;
+    curr.predicate = GEN_PREDICATE_NONE;
+    curr.noMask = 1;
+    curr.quarterControl = GEN_COMPRESSION_Q1;
     MOV(dest, r);
+    if (width == 16) {
+      curr.quarterControl = GEN_COMPRESSION_Q2;
+      MOV(GenRegister::offset(dest, 2), r);
+    }
     pop();
   }
 
@@ -836,17 +883,25 @@ namespace gbe
     MOV(dest, src0);
   }
 
+  void GenEncoder::LOAD_INT64_IMM(GenRegister dest, int64_t value) {
+    GenRegister u0 = GenRegister::immd((int)value), u1 = GenRegister::immd(value >> 32);
+    int execWidth = curr.execWidth;
+    push();
+    curr.execWidth = 8;
+    for(int nib = 0; nib < execWidth/4; nib ++) {
+      curr.chooseNib(nib);
+      MOV(dest.top_half(), u1);
+      MOV(dest.bottom_half(), u0);
+      dest = GenRegister::suboffset(dest, 4);
+    }
+    pop();
+  }
+
   void GenEncoder::MOV_DF(GenRegister dest, GenRegister src0, GenRegister r) {
     int w = curr.execWidth;
     if (src0.isdf()) {
-      push();
-      curr.execWidth = 16;
-      MOV(dest, src0);
-      if (w == 16) {
-        curr.quarterControl = 1;
-        MOV(GenRegister::QnPhysical(dest, w / 4), GenRegister::QnPhysical(src0, w / 4));
-      }
-      pop();
+      GBE_ASSERT(0); // MOV DF is called from convert instruction,
+                     // We should never convert a df to a df.
     } else {
       GenRegister r0 = GenRegister::h2(r);
       push();
@@ -903,6 +958,13 @@ namespace gbe
   ALU2(MACH)
   ALU3(MAD)
 
+  void GenEncoder::SUBB(GenRegister dest, GenRegister src0, GenRegister src1) {
+    push();
+    curr.accWrEnable = 1;
+    alu2(this, GEN_OPCODE_SUBB, dest, src0, src1);
+    pop();
+  }
+
   void GenEncoder::ADDC(GenRegister dest, GenRegister src0, GenRegister src1) {
     push();
     curr.accWrEnable = 1;
@@ -1117,15 +1179,16 @@ namespace gbe
                           bool header_present,
                           unsigned char bti,
                           unsigned char sampler,
+                          unsigned int coord_cnt,
                           uint32_t simdWidth,
                           uint32_t writemask,
                           uint32_t return_format)
   {
      if (writemask == 0) return;
-     uint32_t msg_type = (simdWidth == 16) ?
+     uint32_t msg_type =  (simdWidth == 16) ?
                             GEN_SAMPLER_MESSAGE_SIMD16_SAMPLE : GEN_SAMPLER_MESSAGE_SIMD8_SAMPLE;
      uint32_t response_length = (4 * (simdWidth / 8));
-     uint32_t msg_length = (2 * (simdWidth / 8));
+     uint32_t msg_length = (coord_cnt * (simdWidth / 8));
      if (header_present)
        msg_length++;
      uint32_t simd_mode = (simdWidth == 16) ?
@@ -1150,6 +1213,49 @@ namespace gbe
      this->setSrc0(insn, msg);
      setTypedWriteMessage(this, insn, bti, msg_type, msg_length, header_present);
   }
+  static void setScratchMessage(GenEncoder *p,
+                                   GenInstruction *insn,
+                                   uint32_t offset,
+                                   uint32_t block_size,
+                                   uint32_t channel_mode,
+                                   uint32_t msg_type,
+                                   uint32_t msg_length,
+                                   uint32_t response_length)
+  {
+     const GenMessageTarget sfid = GEN_SFID_DATAPORT_DATA_CACHE;
+     setMessageDescriptor(p, insn, sfid, msg_length, response_length, true);
+     insn->bits3.gen7_scratch_rw.block_size = block_size;
+     insn->bits3.gen7_scratch_rw.msg_type = msg_type;
+     insn->bits3.gen7_scratch_rw.channel_mode = channel_mode;
+     insn->bits3.gen7_scratch_rw.offset = offset;
+     insn->bits3.gen7_scratch_rw.category = 1;
+  }
+
+  void GenEncoder::SCRATCH_WRITE(GenRegister msg, uint32_t offset, uint32_t size, uint32_t src_num, uint32_t channel_mode)
+  {
+     assert(src_num == 1 || src_num ==2);
+     uint32_t block_size = src_num == 1 ? GEN_SCRATCH_BLOCK_SIZE_1 : GEN_SCRATCH_BLOCK_SIZE_2;
+     GenInstruction *insn = this->next(GEN_OPCODE_SEND);
+     this->setHeader(insn);
+     this->setDst(insn, GenRegister::retype(GenRegister::null(), GEN_TYPE_UD));
+     this->setSrc0(insn, msg);
+     this->setSrc1(insn, GenRegister::immud(0));
+     // here src_num means register that will be write out: in terms of 32byte register number
+     setScratchMessage(this, insn, offset, block_size, channel_mode, GEN_SCRATCH_WRITE, src_num+1, 0);
+  }
+
+  void GenEncoder::SCRATCH_READ(GenRegister dst, GenRegister src, uint32_t offset, uint32_t size, uint32_t dst_num, uint32_t channel_mode)
+  {
+     assert(dst_num == 1 || dst_num ==2);
+     uint32_t block_size = dst_num == 1 ? GEN_SCRATCH_BLOCK_SIZE_1 : GEN_SCRATCH_BLOCK_SIZE_2;
+     GenInstruction *insn = this->next(GEN_OPCODE_SEND);
+     this->setHeader(insn);
+     this->setDst(insn, dst);
+     this->setSrc0(insn, src);
+     this->setSrc1(insn, GenRegister::immud(0));
+      // here dst_num is the register that will be write-back: in terms of 32byte register
+     setScratchMessage(this, insn, offset, block_size, channel_mode, GEN_SCRATCH_READ, 1, dst_num);
+  }
 
   void GenEncoder::EOT(uint32_t msg) {
     GenInstruction *insn = this->next(GEN_OPCODE_SEND);
