@@ -29,9 +29,12 @@
 
 #include <stdint.h>
 #include <stdlib.h>
+#include <semaphore.h>
 
 /*! This will make the compiler output Gen ISA code */
 extern void genSetupCallBacks(void);
+extern sem_t llvm_semaphore;
+extern void genSetupLLVMSemaphore(void);
 
 #endif /* __GBE_GEN_PROGRAM_H__ */
 
