@@ -31,6 +31,8 @@
 #include <algorithm>
 #include <climits>
 
+#define RESERVED_REG_NUM_FOR_SPILL 6
+
 namespace gbe
 {
   /////////////////////////////////////////////////////////////////////////////
@@ -63,8 +65,10 @@ namespace gbe
     void allocateFlags(Selection &selection);
     /*! Allocate the GRF registers */
     bool allocateGRFs(Selection &selection);
+    /*! Create gen registers for all preallocated curbe registers. */
+    void allocatePayloadRegs(void);
     /*! Create a Gen register from a register set in the payload */
-    void allocatePayloadReg(gbe_curbe_type, ir::Register, uint32_t subValue = 0, uint32_t subOffset = 0);
+    void allocatePayloadReg(ir::Register, uint32_t offset, uint32_t subOffset = 0);
     /*! Create the intervals for each register */
     /*! Allocate the vectors detected in the instruction selection pass */
     void allocateVector(Selection &selection);
@@ -84,8 +88,6 @@ namespace gbe
     map<ir::Register, VectorLocation> vectorMap;
     /*! All vectors used in the selection */
     vector<SelectionVector*> vectors;
-    /*! All vectors that are already expired */
-    set<SelectionVector*> expired;
     /*! The set of booleans that will go to GRF (cannot be kept into flags) */
     set<ir::Register> grfBooleans;
     /*! All the register intervals */
@@ -94,6 +96,10 @@ namespace gbe
     vector<GenRegInterval*> starting;
     /*! Intervals sorting based on ending point positions */
     vector<GenRegInterval*> ending;
+    /*! registers that are spilled */
+    set<ir::Register> spilled;
+    /* reserved registers for register spill/reload */
+    uint32_t reservedReg;
     /*! Current vector to expire */
     uint32_t expiringID;
     /*! Use custom allocator */
@@ -118,18 +124,39 @@ namespace gbe
   GenRegAllocator::Opaque::Opaque(GenContext &ctx) : ctx(ctx) {}
   GenRegAllocator::Opaque::~Opaque(void) {}
 
-  void GenRegAllocator::Opaque::allocatePayloadReg(gbe_curbe_type value,
-                                                   ir::Register reg,
-                                                   uint32_t subValue,
+  void GenRegAllocator::Opaque::allocatePayloadReg(ir::Register reg,
+                                                   uint32_t offset,
                                                    uint32_t subOffset)
   {
     using namespace ir;
-    const Kernel *kernel = ctx.getKernel();
-    const int32_t curbeOffset = kernel->getCurbeOffset(value, subValue);
-    if (curbeOffset >= 0) {
-      const uint32_t offset = GEN_REG_SIZE + curbeOffset + subOffset;
-      RA.insert(std::make_pair(reg, offset));
-      this->intervals[reg].minID = 0;
+    assert(offset >= GEN_REG_SIZE);
+    offset += subOffset;
+    RA.insert(std::make_pair(reg, offset));
+    GBE_ASSERT(reg != ocl::blockip || (offset % GEN_REG_SIZE == 0));
+    this->intervals[reg].minID = 0;
+    this->intervals[reg].maxID = 0;
+  }
+
+  INLINE void GenRegAllocator::Opaque::allocatePayloadRegs(void) {
+    using namespace ir;
+    for(auto &it : this->ctx.curbeRegs)
+      if (it.first.value() < 0x8000)
+        allocatePayloadReg(it.first, it.second);
+
+    // Allocate all pushed registers (i.e. structure kernel arguments)
+    const Function &fn = ctx.getFunction();
+    GBE_ASSERT(fn.getProfile() == PROFILE_OCL);
+    const Function::PushMap &pushMap = fn.getPushMap();
+    for (auto rit = pushMap.rbegin(); rit != pushMap.rend(); ++rit) {
+      const uint32_t argID = rit->second.argID;
+      const FunctionArgument arg = fn.getArg(argID);
+
+      const uint32_t subOffset = rit->second.offset;
+      const Register reg = rit->second.getRegister();
+      auto it = this->ctx.curbeRegs.find(arg.reg);
+      assert(it != ctx.curbeRegs.end());
+      allocatePayloadReg(reg, it->second, subOffset);
+      ctx.splitBlock(it->second, subOffset);
     }
   }
 
@@ -259,6 +286,11 @@ namespace gbe
         continue;
       }
 
+      //ignore register that already spilled
+      if(spilled.contains(reg)) {
+        this->expiringID++;
+        continue;
+      }
       // Ignore booleans that were allocated with flags
       // if (ctx.getRegisterFamily(reg) == ir::FAMILY_BOOL && !grfBooleans.contains(reg)) {
       if (ctx.sel->getRegisterFamily(reg) == ir::FAMILY_BOOL) {
@@ -276,28 +308,9 @@ namespace gbe
         continue;
       }
       // Case 1 - it does not belong to a vector. Just remove it
-      if (vectorMap.contains(reg) == false) {
         ctx.deallocate(it->second);
         this->expiringID++;
         return true;
-      // Case 2 - check that the vector has not been already removed. If not,
-      // since we equaled the intervals of all registers in the vector, we just
-      // remove the complete vector
-      } else {
-        SelectionVector *vector = vectorMap.find(reg)->second.first;
-        if (expired.contains(vector)) {
-          this->expiringID++;
-          continue;
-        } else {
-          const ir::Register first = vector->reg[0].reg();
-          auto it = RA.find(first);
-          GBE_ASSERT(it != RA.end());
-          ctx.deallocate(it->second);
-          expired.insert(vector);
-          this->expiringID++;
-          return true;
-        }
-      }
     }
 
     // We were not able to expire anything
@@ -445,7 +458,7 @@ namespace gbe
 
           // Compare instructions update the flags so we must copy it back to
           // the GRF
-          if (insn.opcode == SEL_OP_CMP) {
+          if (insn.opcode == SEL_OP_CMP || insn.opcode == SEL_OP_I64CMP) {
             SelectionInstruction *mov1 = selection.create(SEL_OP_MOV,1,1);
             mov1->state = mov0->state;
             mov1->dst(0) = mov0->src(0);
@@ -473,33 +486,66 @@ namespace gbe
       auto it = vectorMap.find(reg);
       if (it != vectorMap.end()) {
         const SelectionVector *vector = it->second.first;
+        // all the reg in the SelectionVector are spilled
+        if(spilled.contains(vector->reg[0].reg()))
+          continue;
         const uint32_t simdWidth = ctx.getSimdWidth();
-        const uint32_t alignment = simdWidth * sizeof(uint32_t);
+
+        const ir::RegisterData regData = ctx.sel->getRegisterData(reg);
+        const ir::RegisterFamily family = regData.family;
+        const uint32_t typeSize = familyVectorSize[family];
+        const uint32_t alignment = simdWidth*typeSize;
+
         const uint32_t size = vector->regNum * alignment;
+
         uint32_t grfOffset;
         while ((grfOffset = ctx.allocate(size, alignment)) == 0) {
           const bool success = this->expireGRF(interval);
-          if (success == false) return false;
+          if (success == false) {
+            // if no spill support, just return false, else simply spill the register
+            if(reservedReg == 0) return false;
+            break;
+          }
         }
-        for (uint32_t regID = 0; regID < vector->regNum; ++regID, grfOffset += alignment) {
+        if(grfOffset == 0) {
+          // spill all the registers in the SelectionVector
+          // the tricky here is I need to use reservedReg+1 as scratch write payload.
+          // so, i need to write the first register to scratch memory first.
+          // the spillReg() will just append scratch write insn after the def. To spill
+          // the first register, need to call spillReg() last for the vector->reg[0]
+          GBE_ASSERT(vector->regNum < RESERVED_REG_NUM_FOR_SPILL);
+          for(int i = vector->regNum-1; i >= 0; i--) {
+            spilled.insert(vector->reg[i].reg());
+            selection.spillReg(vector->reg[i].reg(), reservedReg);
+          }
+          continue;
+        }
+        for (uint32_t regID = 0; regID < vector->regNum; ++regID) {
           const ir::Register reg = vector->reg[regID].reg();
-          GBE_ASSERT(RA.contains(reg) == false);
-          RA.insert(std::make_pair(reg, grfOffset));
+          GBE_ASSERT(RA.contains(reg) == false
+                     && ctx.sel->getRegisterData(reg).family == family);
+          RA.insert(std::make_pair(reg, grfOffset + alignment * regID));
+          ctx.splitBlock(grfOffset, alignment * regID);  //splitBlock will not split if regID == 0
         }
       }
       // Case 2: This is a regular scalar register, allocate it alone
-      else if (this->createGenReg(interval) == false)
-        return false;
+      else if (this->createGenReg(interval) == false) {
+        if(reservedReg == 0) return false;
+        spilled.insert(reg);
+        selection.spillReg(reg, reservedReg);
+      }
     }
     return true;
   }
 
   INLINE bool GenRegAllocator::Opaque::allocate(Selection &selection) {
     using namespace ir;
-    const Kernel *kernel = ctx.getKernel();
-    const Function &fn = ctx.getFunction();
-    GBE_ASSERT(fn.getProfile() == PROFILE_OCL);
-
+    if (ctx.getSimdWidth() == 8) {
+      reservedReg = ctx.allocate(RESERVED_REG_NUM_FOR_SPILL * GEN_REG_SIZE, GEN_REG_SIZE);
+      reservedReg /= GEN_REG_SIZE;
+    } else {
+      reservedReg = 0;
+    }
     // Allocate all the vectors first since they need to be contiguous
     this->allocateVector(selection);
     // schedulePreRegAllocation(ctx, selection);
@@ -509,25 +555,7 @@ namespace gbe
       this->intervals.push_back(ir::Register(regID));
 
     // Allocate the special registers (only those which are actually used)
-    allocatePayloadReg(GBE_CURBE_LOCAL_ID_X, ocl::lid0);
-    allocatePayloadReg(GBE_CURBE_LOCAL_ID_Y, ocl::lid1);
-    allocatePayloadReg(GBE_CURBE_LOCAL_ID_Z, ocl::lid2);
-    allocatePayloadReg(GBE_CURBE_LOCAL_SIZE_X, ocl::lsize0);
-    allocatePayloadReg(GBE_CURBE_LOCAL_SIZE_Y, ocl::lsize1);
-    allocatePayloadReg(GBE_CURBE_LOCAL_SIZE_Z, ocl::lsize2);
-    allocatePayloadReg(GBE_CURBE_GLOBAL_SIZE_X, ocl::gsize0);
-    allocatePayloadReg(GBE_CURBE_GLOBAL_SIZE_Y, ocl::gsize1);
-    allocatePayloadReg(GBE_CURBE_GLOBAL_SIZE_Z, ocl::gsize2);
-    allocatePayloadReg(GBE_CURBE_GLOBAL_OFFSET_X, ocl::goffset0);
-    allocatePayloadReg(GBE_CURBE_GLOBAL_OFFSET_Y, ocl::goffset1);
-    allocatePayloadReg(GBE_CURBE_GLOBAL_OFFSET_Z, ocl::goffset2);
-    allocatePayloadReg(GBE_CURBE_WORK_DIM, ocl::workdim);
-    allocatePayloadReg(GBE_CURBE_GROUP_NUM_X, ocl::numgroup0);
-    allocatePayloadReg(GBE_CURBE_GROUP_NUM_Y, ocl::numgroup1);
-    allocatePayloadReg(GBE_CURBE_GROUP_NUM_Z, ocl::numgroup2);
-    allocatePayloadReg(GBE_CURBE_STACK_POINTER, ocl::stackptr);
-    allocatePayloadReg(GBE_CURBE_THREAD_NUM, ocl::threadn);
-    allocatePayloadReg(GBE_CURBE_GLOBAL_CONSTANT_OFFSET, ocl::constoffst);
+    this->allocatePayloadRegs();
 
     // Group and barrier IDs are always allocated by the hardware in r0
     RA.insert(std::make_pair(ocl::groupid0,  1*sizeof(float))); // r0.1
@@ -536,33 +564,6 @@ namespace gbe
     RA.insert(std::make_pair(ocl::barrierid, 2*sizeof(float))); // r0.2
 
     // block IP used to handle the mask in SW is always allocated
-    const int32_t blockIPOffset = GEN_REG_SIZE + kernel->getCurbeOffset(GBE_CURBE_BLOCK_IP,0);
-    GBE_ASSERT(blockIPOffset >= 0 && blockIPOffset % GEN_REG_SIZE == 0);
-    RA.insert(std::make_pair(ocl::blockip, blockIPOffset));
-    this->intervals[ocl::blockip].minID = 0;
-
-    // Allocate all (non-structure) argument parameters
-    const uint32_t argNum = fn.argNum();
-    for (uint32_t argID = 0; argID < argNum; ++argID) {
-      const FunctionArgument &arg = fn.getArg(argID);
-      GBE_ASSERT(arg.type == FunctionArgument::GLOBAL_POINTER ||
-                 arg.type == FunctionArgument::CONSTANT_POINTER ||
-                 arg.type == FunctionArgument::LOCAL_POINTER ||
-                 arg.type == FunctionArgument::VALUE ||
-                 arg.type == FunctionArgument::STRUCTURE ||
-                 arg.type == FunctionArgument::IMAGE ||
-                 arg.type == FunctionArgument::SAMPLER);
-      allocatePayloadReg(GBE_CURBE_KERNEL_ARGUMENT, arg.reg, argID);
-    }
-
-    // Allocate all pushed registers (i.e. structure kernel arguments)
-    const Function::PushMap &pushMap = fn.getPushMap();
-    for (const auto &pushed : pushMap) {
-      const uint32_t argID = pushed.second.argID;
-      const uint32_t subOffset = pushed.second.offset;
-      const Register reg = pushed.second.getRegister();
-      allocatePayloadReg(GBE_CURBE_KERNEL_ARGUMENT, reg, argID, subOffset);
-    }
 
     // Compute the intervals
     int32_t insnID = 0;
@@ -599,7 +600,8 @@ namespace gbe
 
         // Flag registers can only go to src[0]
         const SelectionOpcode opcode = SelectionOpcode(insn.opcode);
-        if (opcode == SEL_OP_AND || opcode == SEL_OP_OR) {
+        if (opcode == SEL_OP_AND || opcode == SEL_OP_OR || opcode == SEL_OP_XOR
+            || opcode == SEL_OP_I64AND || opcode == SEL_OP_I64OR || opcode == SEL_OP_I64XOR) {
           if (insn.src(1).physical == 0) {
             const ir::Register reg = insn.src(1).reg();
             if (ctx.sel->getRegisterFamily(reg) == ir::FAMILY_BOOL)
@@ -627,27 +629,6 @@ namespace gbe
       }
     }
 
-    // Extend the liveness of the registers that belong to vectors. Actually,
-    // this is way too brutal, we should instead maintain a list of allocated
-    // intervals to handle vector registers independently while doing the linear
-    // scan (or anything else)
-    for (auto vector : this->vectors) {
-      const uint32_t regNum = vector->regNum;
-      const ir::Register first = vector->reg[0].reg();
-      int32_t minID = this->intervals[first].minID;
-      int32_t maxID = this->intervals[first].maxID;
-      for (uint32_t regID = 1; regID < regNum; ++regID) {
-        const ir::Register reg = vector->reg[regID].reg();
-        minID = std::min(minID, this->intervals[reg].minID);
-        maxID = std::max(maxID, this->intervals[reg].maxID);
-      }
-      for (uint32_t regID = 0; regID < regNum; ++regID) {
-        const ir::Register reg = vector->reg[regID].reg();
-        this->intervals[reg].minID = minID;
-        this->intervals[reg].maxID = maxID;
-      }
-    }
-
     // Sort both intervals in starting point and ending point increasing orders
     const uint32_t regNum = ctx.sel->getRegNum();
     this->starting.resize(regNum);
@@ -684,6 +665,10 @@ namespace gbe
         int subreg = offst % 8;
         std::cout << "%" << vReg << " g" << reg << "." << subreg << "D" << std::endl;
     }
+    std::set<ir::Register>::iterator is;
+    std::cout << "## spilled registers:" << std::endl;
+    for(is = spilled.begin(); is != spilled.end(); is++)
+      std::cout << (int)*is << std::endl;
     std::cout << std::endl;
   }
 
@@ -698,9 +683,13 @@ namespace gbe
 
   INLINE GenRegister GenRegAllocator::Opaque::genReg(const GenRegister &reg) {
     if (reg.file == GEN_GENERAL_REGISTER_FILE) {
+      if(reg.physical == 1) {
+        return reg;
+      }
       GBE_ASSERT(RA.contains(reg.reg()) != false);
       const uint32_t grfOffset = RA.find(reg.reg())->second;
-      const GenRegister dst = setGenReg(reg, grfOffset);
+      const uint32_t suboffset = reg.subphysical ? reg.subnr : 0;
+      const GenRegister dst = setGenReg(reg, grfOffset + suboffset);
       if (reg.quarter != 0)
         return GenRegister::Qn(dst, reg.quarter);
       else
