@@ -70,6 +70,8 @@ namespace gbe
   INLINE int typeSize(uint32_t type) {
     switch(type) {
       case GEN_TYPE_DF:
+      case GEN_TYPE_UL:
+      case GEN_TYPE_L:
         return 8;
       case GEN_TYPE_UD:
       case GEN_TYPE_D:
@@ -134,6 +136,33 @@ namespace gbe
     uint32_t predicate:4;
     uint32_t inversePredicate:1;
     uint32_t saturate:1;
+    void chooseNib(int nib) {
+      switch (nib) {
+        case 0:
+          quarterControl = 0;
+          nibControl = 0;
+          break;
+        case 1:
+          quarterControl = 0;
+          nibControl = 1;
+          break;
+        case 2:
+          quarterControl = 1;
+          nibControl = 0;
+          break;
+        case 3:
+          quarterControl = 1;
+          nibControl = 1;
+          break;
+        default:
+          NOT_IMPLEMENTED;
+      }
+    }
+    void useFlag(int nr, int subnr) {
+      flag = nr;
+      subFlag = subnr;
+      physicalFlag = 1;
+    }
   };
 
   /*! This is a book-keeping structure used to encode both virtual and physical
@@ -200,11 +229,13 @@ namespace gbe
       int32_t d;
       uint32_t ud;
       uint16_t reg;
+      int64_t i64;
     } value;
 
     uint32_t nr:8;         //!< Just for some physical registers (acc, null)
     uint32_t subnr:8;      //!< Idem
     uint32_t physical:1;   //!< 1 if physical, 0 otherwise
+    uint32_t subphysical:1;//!< 1 if subnr is physical, 0 otherwise
     uint32_t type:4;       //!< Gen type
     uint32_t file:2;       //!< Register file
     uint32_t negation:1;   //!< For source
@@ -222,18 +253,53 @@ namespace gbe
       return r;
     }
 
+    INLINE bool isint64(void) const {
+      if ((type == GEN_TYPE_UL || type == GEN_TYPE_L) && file == GEN_GENERAL_REGISTER_FILE)
+        return true;
+      return false;
+    }
+
     INLINE bool isimmdf(void) const {
       if (type == GEN_TYPE_DF && file == GEN_IMMEDIATE_VALUE)
         return true;
       return false;
     }
 
+    INLINE GenRegister top_half(void) const {
+      GenRegister r = bottom_half();
+      r.subnr += 4;
+      return r;
+    }
+
+    INLINE GenRegister bottom_half(void) const {
+      GBE_ASSERT(isint64());
+      GenRegister r = *this;
+      r.type = type == GEN_TYPE_UL ? GEN_TYPE_UD : GEN_TYPE_D;
+      r.hstride = GEN_HORIZONTAL_STRIDE_2;
+      r.vstride = GEN_VERTICAL_STRIDE_16;
+      return r;
+    }
+
+    INLINE bool is_signed_int(void) const {
+      if ((type == GEN_TYPE_B || type == GEN_TYPE_W || type == GEN_TYPE_D || type == GEN_TYPE_L) && file == GEN_GENERAL_REGISTER_FILE)
+        return true;
+      return false;
+    }
+
     INLINE bool isdf(void) const {
       if (type == GEN_TYPE_DF && file == GEN_GENERAL_REGISTER_FILE)
         return true;
       return false;
     }
 
+    INLINE int flag_nr(void) const {
+      return nr & 15;
+    }
+
+    INLINE int flag_subnr(void) const {
+      return subnr / typeSize(type);
+    }
+
     static INLINE GenRegister h2(GenRegister reg) {
       GenRegister r = reg;
       r.hstride = GEN_HORIZONTAL_STRIDE_2;
@@ -412,6 +478,12 @@ namespace gbe
                          GEN_HORIZONTAL_STRIDE_0);
     }
 
+    static INLINE GenRegister immint64(int64_t i) {
+      GenRegister immediate = imm(GEN_TYPE_L);
+      immediate.value.i64 = i;
+      return immediate;
+    }
+
     static INLINE GenRegister immdf(double df) {
       GenRegister immediate = imm(GEN_TYPE_DF);
       immediate.value.df = df;
@@ -553,6 +625,11 @@ namespace gbe
                          GEN_HORIZONTAL_STRIDE_1);
     }
 
+    static INLINE bool isNull(GenRegister reg) {
+      return (reg.file == GEN_ARCHITECTURE_REGISTER_FILE
+              && reg.nr == GEN_ARF_NULL);
+    }
+
     static INLINE GenRegister acc(void) {
       return GenRegister(GEN_ARCHITECTURE_REGISTER_FILE,
                          GEN_ARF_ACCUMULATOR,
@@ -832,6 +909,26 @@ namespace gbe
                          GEN_HORIZONTAL_STRIDE_2);
     }
 
+    static INLINE GenRegister packed_ud(uint32_t nr, uint32_t subnr) {
+      return GenRegister(GEN_GENERAL_REGISTER_FILE,
+                         nr,
+                         subnr,
+                         GEN_TYPE_UD,
+                         GEN_VERTICAL_STRIDE_8,
+                         GEN_WIDTH_4,
+                         GEN_HORIZONTAL_STRIDE_1);
+    }
+
+    static INLINE GenRegister unpacked_ud(uint32_t nr, uint32_t subnr) {
+      return GenRegister(GEN_GENERAL_REGISTER_FILE,
+                         nr,
+                         subnr,
+                         GEN_TYPE_UD,
+                         GEN_VERTICAL_STRIDE_8,
+                         GEN_WIDTH_4,
+                         GEN_HORIZONTAL_STRIDE_2);
+    }
+
     static INLINE GenRegister mask(uint32_t subnr) {
       return uw1(GEN_ARCHITECTURE_REGISTER_FILE, GEN_ARF_MASK, subnr);
     }
