@@ -206,6 +206,13 @@ namespace ir {
     static bool isClassOf(const Instruction &insn);
   };
 
+  /*! Ternary instructions are typed. dst and sources share the same type */
+  class TernaryInstruction : public Instruction {
+   public:
+    Type getType(void) const;
+    static bool isClassOf(const Instruction &insn);
+  };
+
   /*! Select instructions writes src0 to dst if cond is true. Otherwise, it
    *  writes src1
    */
@@ -353,11 +360,12 @@ namespace ir {
 
   typedef union {
     struct {
-     uint16_t index; /*! the allocated image index */
-     uint16_t type;  /*! the information type */
+     uint8_t index; /*! the allocated image index */
+     uint8_t  type;  /*! the information type */
     };
     uint32_t data;
   } ImageInfoKey;
+
   /*! Get image information */
   class GetImageInfoInstruction : public Instruction {
   public:
@@ -392,6 +400,14 @@ namespace ir {
     static bool isClassOf(const Instruction &insn);
   };
 
+  /*! Get image information */
+  class GetSamplerInfoInstruction : public Instruction {
+  public:
+
+    /*! Return true if the given instruction is an instance of this class */
+    static bool isClassOf(const Instruction &insn);
+  };
+
   /*! Branch instruction is the unified way to branch (with or without
    *  predicate)
    */
@@ -519,10 +535,16 @@ namespace ir {
   Instruction SIN(Type type, Register dst, Register src);
   /*! mul_hi.type dst src */
   Instruction MUL_HI(Type type, Register dst, Register src0, Register src1);
+  /*! i64_mul_hi.type dst src */
+  Instruction I64_MUL_HI(Type type, Register dst, Register src0, Register src1);
+  /*! i64madsat.type dst src */
+  Instruction I64MADSAT(Type type, Register dst, Tuple src);
   /*! upsample_short.type dst src */
   Instruction UPSAMPLE_SHORT(Type type, Register dst, Register src0, Register src1);
   /*! upsample_int.type dst src */
   Instruction UPSAMPLE_INT(Type type, Register dst, Register src0, Register src1);
+  /*! upsample_long.type dst src */
+  Instruction UPSAMPLE_LONG(Type type, Register dst, Register src0, Register src1);
   /*! fbh.type dst src */
   Instruction FBH(Type type, Register dst, Register src);
   /*! fbl.type dst src */
@@ -531,6 +553,10 @@ namespace ir {
   Instruction HADD(Type type, Register dst, Register src0, Register src1);
   /*! rhadd.type dst src */
   Instruction RHADD(Type type, Register dst, Register src0, Register src1);
+  /*! i64hadd.type dst src */
+  Instruction I64HADD(Type type, Register dst, Register src0, Register src1);
+  /*! i64rhadd.type dst src */
+  Instruction I64RHADD(Type type, Register dst, Register src0, Register src1);
   /*! tan.type dst src */
   Instruction RCP(Type type, Register dst, Register src);
   /*! abs.type dst src */
@@ -597,6 +623,8 @@ namespace ir {
   Instruction GT(Type type, Register dst, Register src0, Register src1);
   /*! cvt.{dstType <- srcType} dst src */
   Instruction CVT(Type dstType, Type srcType, Register dst, Register src);
+  /*! sat_cvt.{dstType <- srcType} dst src */
+  Instruction SAT_CVT(Type dstType, Type srcType, Register dst, Register src);
   /*! atomic dst addr.space {src1 {src2}} */
   Instruction ATOMIC(AtomicOps opcode, Register dst, AddressSpace space, Tuple src);
   /*! bra labelIndex */
@@ -618,7 +646,9 @@ namespace ir {
   /*! sample textures */
   Instruction SAMPLE(Tuple dst, Tuple src, Type dstType, Type srcType);
   /*! get image information , such as width/height/depth/... */
-  Instruction GET_IMAGE_INFO(int infoType, Tuple dst, Register src);
+  Instruction GET_IMAGE_INFO(int infoType, Register dst, Register src, Register infoReg);
+  /*! get sampler information  */
+  Instruction GET_SAMPLER_INFO(Register dst, Register src);
   /*! label labelIndex */
   Instruction LABEL(LabelIndex labelIndex);
 
